import os
import json
import logging
from bisect import bisect
from datetime import datetime
from logging import config 


from . import setLogger

class LogFileConfig:
    def __init__(self, path, logger, debug):
        self.path = path
        self.log = logger
        self.debug = debug

class LevelFilter(logging.Filter):
    """
    LoggingFilter used to filter one or more specific log levels messages
    """
    def __init__(self, level):
        self.__level = level

    def filter(self, record):
        return record.levelno in self.__level


# Adapted from https://stackoverflow.com/a/68154386
class LevelFormatter(logging.Formatter):
    """LoggingFormatter used to specifiy the formatting per level"""
    def __init__(self, formats: dict[int, str], **kwargs):
        super().__init__()

        if "fmt" in kwargs:
            raise ValueError(
                "Format string must be passed to level-surrogate formatters, "
                "not this one"
            )

        self.formats = sorted(
            (level, logging.Formatter(fmt, **kwargs)) for level, fmt in formats.items()
        )

    def format(self, record: logging.LogRecord) -> str:
        idx = bisect(self.formats, (record.levelno, ), hi=len(self.formats) - 1)
        level, formatter = self.formats[idx]
        return formatter.format(record)


def config_logger(log_file_path, debug=None, disable_existing_loggers=True):
    """
    Function to configure the logging.
    All info is saved in a simple format on the log file path.
    Debug entries are saved to a separate file if debug is True
    Debug and warning and above are save in a verbose format.
    Warning and above are also printed to std.out

    Args:
        log_file_path (str): Folder where all logs for this run are saved
        debug (bool): if true, debug messages are saved
        no_exist_log (bool): if true, existing loggers are disabled
    """
    debug_path = os.path.join(os.path.dirname(log_file_path), "debug.log")
    simple_format = "%(message)s"
    verbose_format = "[%(asctime)s] %(levelname)s [%(filename)s %(name)s %(funcName)s (%(lineno)d)]: %(message)s"  # noqa: E501

    LOGGING_CONFIG = {
        "version": 1,
        "disable_existing_loggers": disable_existing_loggers,
        "formatters":
            {
                "simple_formatter": {
                    "format": simple_format
                },
                "verbose_formatter": {
                    "format": verbose_format
                },
                "bylevel_formatter":
                    {
                        "()": LevelFormatter,
                        "formats":
                            {
                                logging.DEBUG: verbose_format,
                                logging.INFO: simple_format,
                                logging.WARNING: verbose_format,
                            },
                    },
            },
        "filters": {
            "only_debug": {
                "()": LevelFilter,
                "level": [logging.DEBUG]
            }
        },
        "handlers":
            {
                "stream_handler":
                    {
                        "class": "logging.StreamHandler",
                        "formatter": "simple_formatter",
                        "level": "WARNING",
                    },
                "file_handler":
                    {
                        "class": "logging.FileHandler",
                        "formatter": "bylevel_formatter",
                        "filename": log_file_path,
                        "level": "INFO",
                    },
                "file_handler_debug":
                    {
                        "class": "logging.FileHandler",
                        "formatter": "bylevel_formatter",
                        "filename": debug_path,
                        "mode": "w",
                        "delay": True,
                        "filters": ["only_debug"],
                    },
            },
        "loggers":
            {
                None:
                    {
                        "handlers":
                            ["stream_handler", "file_handler", "file_handler_debug"]
                            if debug else ["stream_handler", "file_handler"],
                        "level":
                            "DEBUG",
                    }
            },
    }

    config.dictConfig(LOGGING_CONFIG)


def init_logfile(log, args=None):
    """
    Put some intial information in the logfile

    Args:
        log : Logging instance
        args (dict): Dictionary with all command line arguments
    """
    logging.info(f"Initialize GBMT log file: {log.root.handlers[1].baseFilename} at {datetime.now().strftime('%Y-%m-%d %H:%M:%S')}")
    if args:
        logging.info("Command line arguments:")
        for key, value in args.items():
            logging.info(f"{key}: {value}")
    logging.info("")

def enable_file_logger(
    log_folder: str,
    filename: str,
    debug: bool = False,
    log_name: str | None = None,
    init_data: dict | None = None,
    disable_existing_loggers: bool = False,
):
    """Enable file logging.

    Args:
        log_folder (str): path to the folder where the log file should be stored
        filename (str): name of the log file
        debug (bool): whether to enable debug logging. Defaults to False.
        log_name (str, optional): name of the logger. Defaults to None.
        init_data (dict, optional): initial data to be logged. Defaults to None.
        disable_existing_loggers (bool): whether to disable existing loggers.
    """
    # create log folder if it does not exist
    path = os.path.join(log_folder, filename)
    if not os.path.exists(log_folder):
        os.makedirs(log_folder)

    # configure logging
    config.config_logger(path, debug, disable_existing_loggers=disable_existing_loggers)

    # get logger and init configuration
    log = logging.getLogger(filename) if not log_name else logging.getLogger(log_name)
    log.setLevel(logging.INFO)
    setLogger(log)
    settings = LogFileConfig(path, log, debug)

    # Begin log file
    config.init_logfile(log, json.dumps(init_data, sort_keys=False, indent=2))

    return settings



import numpy as np
import pandas as pd

from pulp import LpAffineExpression, PULP_CBC_CMD, value, LpProblem, LpMinimize, LpVariable
from typing import Callable

from rdkit import Chem, DataStructs
from rdkit.Chem.rdFingerprintGenerator import GetMorganGenerator

from .clustering import MaxMinClustering, RandomClustering
from .logs import logger

class GloballyBalancedSplit:

    """
    Globally balanced splits.

    Attributes
    ----------
    sizes : list[int], optional
        list of sizes of the splits.
    clusters : dict, optional
        dictionary of clusters, where keys are cluster indices and values are indices of molecules.
    clustering_method : Callable, optional
        Clustering method.
    n_splits : int, optional
        Number of splits.
    equal_weight_perc_compounds_as_tasks : bool, optional
        Whether to weight the tasks equally or not.
    absolute_gap : float, optional  
        Absolute gap between the absolute optimal objective and the current one at which the solver stops and returns a solution.
    time_limit_seconds : int, optional
        Time limit for the linear programming problem.
    n_jobs : int, optional
        Number of jobs.
    min_distance : bool, optional
        Whether to calculate the minimum interset distance between the splits.
    """

    def __init__(
            self,
            sizes : list[int] = [0.8, 0.1, 0.1],
            clusters : dict = None,
            clustering_method : Callable | None = MaxMinClustering(),
            n_splits : int = 1,
            equal_weight_perc_compounds_as_tasks : bool = True,
            absolute_gap : float = 1e-3,
            time_limit_seconds : int = None,
            n_jobs : int = 1,
            min_distance : bool = True,  
            stratify : bool = True,
            stratify_reg_nbins : int = 5,  
            ) -> None:     
        
        if clusters is None and clustering_method is None:
            raise ValueError('Either clusters or clustering_method must be provided.')
        elif clusters is not None and clustering_method is not None:
            raise ValueError('Only one of clusters or clustering_method must be provided.')
        
        self.sizes = sizes
        self.clusters = clusters
        self.clustering_method = clustering_method
        self.n_splits = n_splits
        self.equal_weight_perc_compounds_as_tasks = equal_weight_perc_compounds_as_tasks
        self.absolute_gap = absolute_gap
        self.time_limit_seconds = time_limit_seconds
        self.n_jobs = n_jobs
        self.min_distance = min_distance
        self.stratify = stratify
        self.stratify_reg_nbins = stratify_reg_nbins

    def __call__(
            self,
            data : pd.DataFrame,
            smiles_column : str = 'SMILES',
            tasks : list[str] = None,
            ignore_columns : list[str] = None,
            preassigned_smiles : dict[str, int] = None,
            ) -> pd.DataFrame:
        
        """
        Globally balanced splits.

        Parameters
        ----------
        data : pd.DataFrame
            Dataframe with SMILES strings and tasks.
        smiles_column : str, optional
            Name of the column with SMILES strings.
        tasks : list[str], optional
            list of task columns.
        ignore_columns : list[str], optional
            list of columns to ignore.
        
        Returns
        -------
        splits : pd.DataFrame
            Dataframe with the splits.
        """

        # Multiple splits doesn't work with all clustering methods
        if self.n_splits > 1 :
            if self.clusters is not None:
                raise ValueError('n_splits > 1 does not work with precomputed clusters.')
            elif not (isinstance(self.clustering_method, MaxMinClustering)
                                    or isinstance(self.clustering_method, RandomClustering)):
                raise ValueError(f'n_splits > 1 does not work with type \
                                  {type(self.clustering_method).__name__}.')

        self.df = data.copy()
        self.smiles_column = smiles_column
        self.ignore_columns = ignore_columns
        self.preassigned_smiles = preassigned_smiles
        
        # Save the original tasks and create the tasks for balancing
        self._set_original_tasks(tasks)
        self._set_tasks_for_balancing()


        smiles_list = self.df[smiles_column].tolist()

        for split in range(self.n_splits):

            if self.n_splits == 1:
                split_name = 'Split'
                mintd_name = 'minInterSetTd'
            else:
                split_name = f'Split_{split}'
                mintd_name = f'minInterSetTd_{split}'

            txt = f' {split_name} '
            n = int((80 - len(txt)) / 2)
            logger.info('=' * n + txt + '=' * n)
            logger.info(f'Clustering method: {self.clustering_method.get_name() if self.clustering_method else "precomputed clusters"}')
            logger.info(f'Original tasks: {self.original_tasks}')
            logger.info(f'Tasks for balancing: {self.tasks_for_balancing}')
            logger.info(f'Subset sizes: {self.sizes}')

            # Cluster molecules
            if self.clusters :
                clusters = self.clusters
            else:
                clusters = self.clustering_method(smiles_list, )
            logger.info(f'Number of initial clusters: {len(clusters)}')
            
            # Get clusters indices of pre-assigned molecules
            preassigned_clusters = self._get_preassigned_clusters(clusters) if preassigned_smiles else None
            # Compute the number of self.dfpoints per task for each cluster
            tasks_per_cluster = self._compute_tasks_per_cluster(self.tasks_for_balancing, clusters)
                     
            # Merge the clusters with a linear programming method to create the subsets
            merged_clusters_mapping = self._merge_clusters_with_balancing_mapping(
                tasks_per_cluster, 
                self.sizes, 
                self.equal_weight_perc_compounds_as_tasks, 
                self.absolute_gap,
                self.time_limit_seconds if self.time_limit_seconds else self.get_default_time_limit_seconds(len(smiles_list), len(self.tasks_for_balancing)),
                self.n_jobs,
                preassigned_clusters)  

            for i, idx in clusters.items(): 
                self.df.loc[idx, split_name] = merged_clusters_mapping[i]-1

            # Print balance
            self._compute_task_balace(split_name)
        
            # Compute the minimum interset distance between the splits
            if self.min_distance:
                if hasattr(self.clustering_method, 'fp_calculator'):
                    fp_calculator = self.clustering_method.fp_calculator
                    self._compute_min_interset_Tanimoto_distances(split_name, mintd_name, fp_calculator)
                else:
                    self._compute_min_interset_Tanimoto_distances(split_name, mintd_name)

            # # Print balance and chemical bias metrics
            # self._print_metrics(self.df)

            # Update clustering seed
            if self.n_splits > 1:
                self.clustering_method.seed += 1


        # Drop the tasks for balancing
        cols2drop = [col for col in self.tasks_for_balancing if col not in self.original_tasks]
        self.df.drop(cols2drop, axis=1, inplace=True)


        logger.info('=' * 80)

        return self.df
    
    def get_default_time_limit_seconds(self, nmols : int, ntasks : int) -> int:
        """
        Default time limit in seconds for the optimization problem.

        tlim = nmol^(1/3) * sqrt(ntask) and tlim >= 10s and tlim <= 1h

        Parameters
        ----------
        nmols : int
            Number of molecules.
        ntasks : int
            Number of tasks.
        
        Returns
        -------
        int
            Time limit in seconds.
        """
        
        tmol = nmols ** (1/3)
        ttarget = np.sqrt(ntasks)
        tmin = 10
        tmax = 60 * 60
        tlim = min(tmax, max(tmin, tmol * ttarget)) 
        logger.info(f'Time limit: {int(tlim)}s')
        return tlim


    def _compute_tasks_per_cluster(
            self, 
            tasks : list[str],
            clusters : dict[int, list[int]]
        ) -> dict[int, list[str]]:
        
        """
        Compute the number of datapoints per task for each cluster.

        Parameters
        ----------
        tasks : list[str]
            list of tasks
        clusters : dict[int, list[int]]
            dictionary of clusters and list of indices of molecules in the cluster

        Returns
        -------
        np.array of shape (len(tasks)+1, len(clusters))
            Array with each columns correspoding to a cluster and each row to a task
            plus the 1st row for the number of molecules per cluster
        """

        task_vs_clusters = np.zeros((len(tasks)+1, len(clusters)))
        task_vs_clusters[0,:] = [ len(cluster) for cluster in clusters.values() ]

        for i, task in enumerate(tasks):
            for j, indices_per_cluster in clusters.items():
                self.df_per_cluster = self.df.iloc[indices_per_cluster]
                task_vs_clusters[i+1,j] = self.df_per_cluster[task].dropna().shape[0]

        return task_vs_clusters  
        
    def _set_original_tasks(self, tasks : list[str] | None) -> None:
        """
        Set the original tasks.
        
        Parameters
        ----------
        tasks : list[str] | None
            list of task columns.
        """

        if tasks is None:
            columns = self.df.columns.tolist()
            # Drop the SMILES column.
            columns.remove(self.smiles_column)
            # Drop the ignore columns.
            if self.ignore_columns is not None:
                columns = [column for column in columns if column not in self.ignore_columns]
            # Drop columns starting with 'Split' and 'MinIntersetDistance'.
            columns = [column for column in columns if not column.startswith('Split')]
            columns = [column for column in columns if not column.startswith('MinIntersetDistance')]
            # Set original tasks.
            self.original_tasks = columns
        else:
            self.original_tasks = tasks

    def _set_tasks_for_balancing(self) -> None:

        """
        Set the tasks for balancing. 

        If stratify is True, the tasks are used to create a column for each class (in case of classification tasks) and
        bin the data (in case of regression tasks). If stratify is False, the tasks are used as is.
        """

        def is_convertible(value):
            try:
                float(value)
                return True
            except ValueError:
                return False

        self.tasks_for_balancing = []
        if self.stratify:
            for task in self.original_tasks:
                values = self.df[task].dropna().unique()
                values_is_numerical = [is_convertible(value) for value in values]
                
                # Check if contains non-convertible strings
                if not (all(values_is_numerical)): # Some non numerical values

                    if any(values_is_numerical): # But some values are numerical
                        raise ValueError(f'Column {task} contains both numerical and strings values.')
                    else:
                        # Create a task per string
                        for string in values:
                            key = f'{task}_{string}'
                            self.df[key] = (self.df[task] == string).map({True: 1, False: np.nan})
                            self.tasks_for_balancing.append(key)
                        logger.info(f'String classification {task} stratified into {len(self.df[task].dropna().unique())} tasks.')
                # Classification: task per class
                elif all( x % 1 == 0 for x in values):
                    for cls in values:
                        key = f'{task}_{int(cls)}'
                        self.df[key] = (self.df[task] == cls).map({True: 1, False: np.nan})
                        self.tasks_for_balancing.append(key)
                    logger.info(f'Classification task {task} stratified into {len(self.df[task].dropna().unique())} tasks.')
                # Regression: bin data and use as tasks
                else:
                    sorted_values = np.sort(values)
                    bins = np.array_split(sorted_values, self.stratify_reg_nbins)
                    for i, bin in enumerate(bins):
                        key = f'{task}_{bin[0]:.2f}_{bin[-1]:.2f}'
                        self.df[key] = self.df[task].apply(lambda x: x if x in bin else np.nan)
                        self.tasks_for_balancing.append(key)
                    logger.info(f'Regression task {task} stratified into {self.stratify_reg_nbins} tasks.')
        else:
            self.tasks_for_balancing = self.original_tasks

    def _get_preassigned_clusters(
            self, 
            clusters : dict[int, list[int]]) -> dict[int, int]:

        """
        Get the pre-assigned clusters.

        Parameters
        ----------
        clusters : dict[int, list[int]]
            dictionary of clusters and list of indices of molecules in the clusters.
        
        Returns
        -------
        dict[int, int]
            dictionary with cluster indices as keys and subset indices as values.
        """

        preassigned_clusters = {}
        for smi, subset in self.preassigned_smiles.items():
            imol = self.df[self.df[self.smiles_column] == smi].index[0]
            for idx, cluster in clusters.items():
                if imol in cluster:
                    if (idx in preassigned_clusters.keys()) and (subset != preassigned_clusters[idx]):
                        raise ValueError(f'Pre-assigned cluster {idx} is assigned to multiple subsets.')
                    preassigned_clusters[idx] = subset
                    logger.info(f'Cluster {idx} contaning {smi} is preassigned to subset {subset}.')
                    break

        preassigned_clusters
        
        return preassigned_clusters

    def _merge_clusters_with_balancing_mapping(
            self, 
            tasks_vs_clusters_array : np.array,
            sizes : list[float] = [0.9, 0.1, 0.1], 
            equal_weight_perc_compounds_as_tasks : bool = False,
            absolute_gap : float = 1e-3,
            time_limit_seconds : int = 60*60,
            max_N_threads : int = 1,
            preassigned_clusters : dict[int, int] | None = None) -> list[list[int]]:
            """
            Linear programming function needed to balance the self.df while merging clusters.

            Paper: Tricarico et al., Construction of balanced, chemically dissimilar training, validation 
            and test sets for machine learning on molecular self.dfsets, 2022, 
            DOI: https://doi.org/10.26434/chemrxiv-2022-m8l33-v2

            Parameters
            ----------
            tasks_vs_clusters_array : 2D np.array
                - the cross-tabulation of the number of self.df points per cluster, per task.
                - columns represent unique clusters.
                - rows represent tasks, except the first row, which represents the number of records (or compounds).
                - Optionally, instead of the number of self.df points, the provided array may contain the *percentages*
                    of self.df points _for the task across all clusters_ (i.e. each *row*, NOT column, may sum to 1).
                IMPORTANT: make sure the array has 2 dimensions, even if only balancing the number of self.df records,
                    so there is only 1 row. This can be achieved by setting ndmin = 2 in the np.array function.
            sizes : list
                - list of the desired final sizes (will be normalised to fractions internally).
            equal_weight_perc_compounds_as_tasks : bool
                - if True, matching the % records will have the same weight as matching the % self.df of individual tasks.
                - if False, matching the % records will have a weight X times larger than the X tasks.
            absolute_gap : float
                - the absolute gap between the absolute optimal objective and the current one at which the solver
                stops and returns a solution. Can be very useful for cases where the exact solution requires
                far too long to be found to be of any practical use.
            time_limit_seconds : int
                - the time limit in seconds for the solver (by default set to 1 hour)
                - after this time, whatever solution is available is returned
            max_N_threads : int
                - the maximal number of threads to be used by the solver.
                - it is advisable to set this number as high as allowed by the available resources.
            preassigned_clusters : dict
                - a dictionary of the form {cluster_index: ML_subset_index} to force the clusters to be assigned
                    to the ML subsets as specified by the user.
            
            Returns
            ------
            list (of length equal to the number of columns of tasks_vs_clusters_array) of final cluster identifiers
                (integers, numbered from 1 to len(sizes)), mapping each unique initial cluster to its final cluster.
            Example: if sizes == [20, 10, 70], the output will be a list like [3, 3, 1, 2, 1, 3...], where
                '1' represents the final cluster of relative size 20, '2' the one of relative size 10, and '3' the 
                one of relative size 70.
            """

            # Calculate the fractions from sizes

            fractional_sizes = sizes / np.sum(sizes)

            S = len(sizes)

            # Normalise the self.df matrix
            tasks_vs_clusters_array = tasks_vs_clusters_array / tasks_vs_clusters_array.sum(axis = 1, keepdims = True)

            # Find the number of tasks + compounds (M) and the number of initial clusters (N)
            M, N = tasks_vs_clusters_array.shape
            if (S > N):
                errormessage = 'The requested number of new clusters to make ('+ str(S) + ') cannot be larger than the initial number of clusters (' + str(N) + '). Please review.'
                raise ValueError(errormessage)

            # Given matrix A (M x N) of fraction of self.df per cluster, assign each cluster to one of S final ML subsets,
            # so that the fraction of self.df per ML subset is closest to the corresponding fraction_size.
            # The weights on each ML subset (WML, S x 1) are calculated from fractional_sizes harmonic-mean-like.
            # The weights on each task (WT, M x 1) are calculated as requested by the user.
            # In the end: argmin SUM(ABS((A.X-T).WML).WT)
            # where X is the (N x S) binary solution matrix
            # where T is the (M x S) matrix of task fraction sizes (repeat of fractional_sizes)
            # constraint: assign one cluster to one and only one final ML subset
            # i.e. each row of X must sum to 1

            A = np.copy(tasks_vs_clusters_array)

            # Create WT = obj_weights
            if ((M > 1) & (equal_weight_perc_compounds_as_tasks == False)):
                obj_weights = np.array([M-1] + [1] * (M-1))
            else:
                obj_weights = np.array([1] * M)

            obj_weights = obj_weights / np.sum(obj_weights)

            # Create WML
            sk_harmonic = (1 / fractional_sizes) / np.sum(1 / fractional_sizes)

            # Round all values to have only 3 decimals > reduce computational time
            A = np.round(A, 3)
            fractional_sizes = np.round(fractional_sizes, 3)
            obj_weights = np.round(obj_weights, 3)
            sk_harmonic = np.round(sk_harmonic, 3)           

            # Create the pulp model
            prob = LpProblem("Data_balancing", LpMinimize)

            # Create the pulp variables
            # x_names represent clusters, ML_subsets, and are binary variables
            x_names = ['x_'+str(i) for i in range(N * S)]
            x = [LpVariable(x_names[i], lowBound = 0, upBound = 1, cat = 'Integer') for i in range(N * S)]
            # X_names represent tasks, ML_subsets, and are continuous positive variables
            X_names = ['X_'+str(i) for i in range(M * S)]
            X = [LpVariable(X_names[i], lowBound = 0, cat = 'Continuous') for i in range(M * S)]

            # Add the objective to the model

            obj = []
            coeff = []
            for m in range(S):
                for t in range(M):
                    obj.append(X[m*M+t])
                    coeff.append(sk_harmonic[m] * obj_weights[t])

            prob += LpAffineExpression([(obj[i],coeff[i]) for i in range(len(obj)) ])

            # Add the constraints to the model

            # Constraints forcing each cluster to be in one and only one ML_subset
            for c in range(N):
                prob += LpAffineExpression([(x[c+m*N],+1) for m in range(S)]) == 1

            # If preassigned_clusters is pro[int, int]vided, add the constraints to the model to force the clusters
            # to be assigned to the ML subset preassigned_clusters[t]
            if preassigned_clusters:
                for c, subset in preassigned_clusters.items():
                    # prob += LpAffineExpression(x[c+(subset)*N]) == 1
                    prob += x[c+(subset)*N] == 1

            # Constraints related to the ABS values handling, part 1 and 2
            for m in range(S):
                for t in range(M):
                    cs = [c for c in range(N) if A[t,c] != 0]
                    prob += LpAffineExpression([(x[c+m*N],A[t,c]) for c in cs]) - X[t] <= fractional_sizes[m]
                    prob += LpAffineExpression([(x[c+m*N],A[t,c]) for c in cs]) + X[t] >= fractional_sizes[m]

            # Solve the model
            prob.solve(PULP_CBC_CMD(gapAbs = absolute_gap, timeLimit = time_limit_seconds, threads = max_N_threads, msg=False))

            # Extract the solution
            list_binary_solution = [value(x[i]) for i in range(N * S)]
            list_initial_cluster_indices = [(list(range(N)) * S)[i] for i,l in enumerate(list_binary_solution) if l == 1]
            list_final_ML_subsets = [(list((1 + np.repeat(range(S), N)).astype('int64')))[i] for i,l in enumerate(list_binary_solution) if l == 1]
            mapping = [x for _, x in sorted(zip(list_initial_cluster_indices, list_final_ML_subsets))]

            return mapping

    def _compute_min_interset_Tanimoto_distances(
            self, 
            split_col = 'Split',
            mTd_col = 'minInterSetTd',
            fp_calculator = GetMorganGenerator(radius=3, fpSize=2048)
            ):
        """
        Compute the minimum Tanimoto distance per compound to the compounds in the other subsets.
        
        Parameters
        ----------
        split_col : str, optional
            Name of the column containing the subset number, by default 'Split'
        mTd_col : str, optional
            Name of the column to store the minimum Tanimoto distance in, by default 'minInterSetTd'
        fp_calculator : rdkit.Chem.rdMolDescriptors.GetMorganGenerator, optional
            Fingerprint calculator, by default GetMorganGenerator(radius=3, nBits=2048)
        """

        # Print header
        txt = ' Min. inter-set Tanimoto distance '
        n = int((80 - len(txt)) / 2)
        logger.info('-' * n + txt + '-' * n) 

        # Compute fingerprints
        mols = [ Chem.MolFromSmiles(smi) for smi in self.df[self.smiles_column].tolist() ]
        fps = [ fp_calculator.GetFingerprint(mol) for mol in mols ]

        min_distances = np.zeros(len(fps))
        # Iterate over subsets and compute minimum Tanimoto distance per compound 
        # to the compounds in the other subsets
        for j in self.df[split_col].unique():
            ref_idx = self.df[self.df[split_col] == j].index.tolist()
            other_fps = [ fp for i, fp in enumerate(fps) if i not in ref_idx ]
            for i in ref_idx :
                sims = DataStructs.BulkTanimotoSimilarity(fps[i], other_fps)
                min_distances[i] = 1. - max(sims)

        self.df[mTd_col] = min_distances

        # Print average and std  of minimum distances per subset
        for subset in sorted(self.df[split_col].unique()):
            dist = self.df[self.df[split_col] == subset][mTd_col] #.to_numpy()
            logger.info(f'Subset {int(subset)}: {dist.mean():.2f} +/- {dist.std():.2f} | {dist.median():.2f}')

        # Chemical dissimilarity score
        cd_score = self.df.groupby(split_col)[mTd_col].median().min()
        logger.info(f'Chemical dissimilarity score: {cd_score:.2f}')
    
    def _compute_task_balace(self, split_col : str = 'Split'):

        """
        Compute the task balance per subset.
        
        Parameters
        ----------
        split_col : str, optional
            Name of the column containing the subset number, by default 'Split'
        """

        # Header
        txt = f' {split_col} balance '
        n = int((80 - len(txt)) / 2)
        logger.info('-' * n + txt + '-' * n) 


        # 1. Print out for each balancing task the fraction and number of self.df points per subset
        # 2. Print out for each original task the fraction and number of self.df points per subset
        # 3. Print the fraction and number of self.df point per subset for all tasks combined

        # Get name of longets task
        longest_task = max(self.tasks_for_balancing, key=len)   

        # 1. Print out for each balancing task the fraction and number of self.df points per subset
        if self.stratify:
            for task in self.tasks_for_balancing:
                txt = f'{task} balance:'
                txt += ' ' * (len(longest_task) - len(task)) + '\t'
                counts = self.df[[task, split_col]].groupby(split_col).count()
                n = counts[task].sum()
                for subset in sorted(self.df[split_col].unique()):
                    n_subset = counts.loc[subset, task]
                    txt += f' {int(subset)}: {n_subset/n:.2f} [{n_subset}]\t'
                logger.info(txt)
            logger.info('')

        # 2. Print out for each original task the fraction and number of self.df points per subset
        for task in self.original_tasks:
            txt = f'{task} balance:'
            txt += ' ' * (len(longest_task) - len(task)) + '\t'
            counts = self.df[[task, split_col]].groupby(split_col).count()
            n = counts[task].sum()
            for subset in sorted(self.df[split_col].unique()):
                n_subset = counts.loc[subset, task]
                txt += f' {int(subset)}: {n_subset/n:.2f} [{n_subset}]\t'
            logger.info(txt)
        logger.info('')

        # 3. Print the fraction and number of self.df point per subset for all tasks combined
        txt = 'Overall balance:' + ' ' * (len(longest_task) - len(task)) + '\t'
        n = self.df.shape[0]
        balance_score = 0
        subsets = sorted(self.df[split_col].unique())
        for i, subset in enumerate(subsets):
            n_subset = self.df[self.df[split_col] == subset].shape[0]
            txt += f' {int(subset)}: {n_subset/n:.2f} [{n_subset}]\t'
            balance_score += np.abs(n_subset/n - self.sizes[i])
        logger.info(txt)
        logger.info(f'Balance score: {balance_score/len(self.sizes):.4f}')


# Unit test for gbmt-splits
import os
import pandas as pd
from unittest import TestCase
from parameterized import parameterized

import logging

from rdkit.Chem.rdFingerprintGenerator import GetMorganGenerator

from gbmtsplits.split import GloballyBalancedSplit
from gbmtsplits.clustering import RandomClustering, MaxMinClustering, LeaderPickerClustering, MurckoScaffoldClustering

preassigned_smiles = {
    'Brc1cccc(Nc2nc3c(N4CCCC4)ncnc3s2)c1' : 0,
    'C#CCn1c(=O)c2c(nc3n2CCCN3C2CCC2)n(C)c1=O' : 1,
}

logging.basicConfig(level=logging.DEBUG)


class TestSplits(TestCase):

    test_data_path = os.path.join(os.path.dirname(__file__), 'test_data.csv')
    seed = 2022
    time_limit = 10

    @parameterized.expand([
        ([0.9, 0.1], None,), 
        ([0.9, 0.1], preassigned_smiles,),
        ])

    def test_random_split(self, sizes, preassigned_smiles):

        data = pd.read_csv(self.test_data_path)
        ncols = data.shape[1]
        clustering = RandomClustering(seed=self.seed)
        splitter = GloballyBalancedSplit(
            sizes = sizes,
            clustering_method = clustering,
            time_limit_seconds=self.time_limit,
        )
        data = splitter(data, preassigned_smiles=preassigned_smiles)

        assert data.shape[1] == ncols + 2
        assert data.Split.nunique() == 2

        if preassigned_smiles is not None:
            for smiles, subset in preassigned_smiles.items():
                assert data.loc[data['SMILES'] == smiles, 'Split'].values[0] == subset


    @parameterized.expand([
        ([0.7, 0.1, 0.1, 0.1], None,), 
        ([0.7, 0.1, 0.1, 0.1], preassigned_smiles,),
        ])
    
    def test_dissimilarity_maxmin_split(self, sizes, preassigned_smiles):
            
        data = pd.read_csv(self.test_data_path)
        ncols = data.shape[1]
        clustering = MaxMinClustering(seed=self.seed)
        splitter = GloballyBalancedSplit(
            sizes = sizes, 
            clustering_method = clustering,
            time_limit_seconds=self.time_limit,
        )
        data = splitter(data, preassigned_smiles=preassigned_smiles)

        assert data.shape[1] == ncols + 2
        assert data.Split.nunique() == 4

        if preassigned_smiles is not None:
            for smiles, subset in preassigned_smiles.items():
                assert data.loc[data['SMILES'] == smiles, 'Split'].values[0] == subset

    def test_dissimilarity_leader_split(self):
            
        data = pd.read_csv(self.test_data_path)
        ncols = data.shape[1]
        clustering = LeaderPickerClustering(
            fp_calculator=GetMorganGenerator(radius=2, fpSize=1024),
            similarity_threshold=0.6)
        splitter = GloballyBalancedSplit(
            clustering_method = clustering,
            time_limit_seconds=self.time_limit,
        )
        data = splitter(data)

        assert data.shape[1] == ncols + 2
        assert data.Split.nunique() == 3

    def test_murcko_scaffold_split(self):
    
        data = pd.read_csv(self.test_data_path)
        ncols = data.shape[1]
        clustering = MurckoScaffoldClustering()
        splitter = GloballyBalancedSplit(
            clustering_method = clustering,
            time_limit_seconds=self.time_limit,
            min_distance=False
        )
        data = splitter(data)

        assert data.shape[1] == ncols + 1
        assert data.Split.nunique() == 3

    def test_multiple_random_splits(self):

        data = pd.read_csv(self.test_data_path)
        ncols = data.shape[1]
        clustering = RandomClustering(seed=self.seed)
        splitter = GloballyBalancedSplit(
            clustering_method = clustering,
            time_limit_seconds=self.time_limit,
            n_splits=3,
            absolute_gap = 1e-3,
        )
        data = splitter(data)

        assert data.shape[1] == ncols + (3 * 2)
        for i in range(3):
            assert data[f'Split_{i}'].nunique() == 3


import os

from .split import GloballyBalancedSplit
from .clustering import *

__version__ = '0.0.5'
if os.path.exists(os.path.join(os.path.dirname(__file__), '_version.py')):
    from ._version import version
    __version__ = version

VERSION = __version__


import numpy as np

from typing import List
from abc import ABC, abstractmethod

from typing import Callable

from rdkit import Chem, DataStructs
from rdkit.Chem.Scaffolds import MurckoScaffold
from rdkit.SimDivFilters import rdSimDivPickers
from rdkit.Chem.rdFingerprintGenerator import GetMorganGenerator

class ClusteringMethod(ABC):

    """
    Abstract base class for clustering methods.
    """

    @abstractmethod
    def __call__(self, smiles_list : List[str]) -> dict:
        pass

    def get_name(self) -> str:
        return self.__class__.__name__

    def _set_n_clusters(self, N : int) -> None:
        self.n_clusters = self.n_clusters if self.n_clusters is not None else N // 10 



class RandomClustering(ClusteringMethod):

    """
    Randomly cluster a list of SMILES strings into n_clusters clusters.
    
    Attributes
    ----------
    n_clusters : int, optional
        Number of clusters.
    seed : int, optional
        Random seed.
    """

    def __init__(self, n_clusters : int = None, seed : int = 42) -> None:
        super().__init__()
        self.n_clusters = n_clusters
        self.seed = seed

    def __call__(self, smiles_list : List[str]) -> dict:
        """
        Randomly cluster a list of SMILES strings into n_clusters clusters.
        
        Parameters
        ----------
        smiles_list : list[str]
            List of SMILES strings to cluster.
        
        Returns
        -------
        clusters : dict
            Dictionary of clusters, where keys are cluster indices and values are indices of SMILES strings.
        """

        self._set_n_clusters(len(smiles_list))

        # Initialize clusters
        clusters = { i: [] for i in range(self.n_clusters) }

        # Randomly assign each molecule to a cluster
        indices = np.random.RandomState(seed=self.seed).permutation(len(smiles_list))
        for i, index in enumerate(indices):
            clusters[i % self.n_clusters].append(index)

        return clusters

        
class MurckoScaffoldClustering(ClusteringMethod):

    """
    Cluster a list of SMILES strings based on Murcko scaffolds.
    """

    def __init__(self) -> None:
        super().__init__()

    def __call__(self, smiles_list : List[str]) -> dict:

        """
        Cluster a list of SMILES strings based on Murcko scaffolds.

        Parameters
        ----------
        smiles_list : list[str]
            List of SMILES strings to cluster.
        """
            
        # Generate scaffolds for each molecule
        mols = [Chem.MolFromSmiles(smiles) for smiles in smiles_list]
        scaffolds = [ MurckoScaffold.GetScaffoldForMol(mol) for mol in mols ]

        # Get unique scaffolds and initialize clusters
        unique_scaffolds = list(set(scaffolds))
        clusters = { i: [] for i in range(len(unique_scaffolds)) }

        # Cluster molecules based on scaffolds
        for i, scaffold in enumerate(scaffolds):
            clusters[unique_scaffolds.index(scaffold)].append(i)

        return clusters
    
class DissimilarityClustering(ClusteringMethod):
    
        """
        Abstract base class for clustering methods based on molecular dissimilarity.
        """
    
        def __init__(self, fp_calculator : Callable = GetMorganGenerator(radius=3, fpSize=2048) ) -> None:
            super().__init__()
            self.fp_calculator = fp_calculator

        def __call__(self, smiles_list : List[str]) -> dict:

            """
            Cluster a list of SMILES strings based on molecular dissimilarity.
            
            Parameters
            ----------
            smiles_list : list[str]
                List of SMILES strings to cluster.
            
            Returns
            -------
            clusters : dict
                Dictionary of clusters, where keys are cluster indices and values are indices of SMILES strings.
            """

            fps = [self.fp_calculator.GetFingerprint(Chem.MolFromSmiles(smiles)) for smiles in smiles_list]

            # Get cluster centroids and initialize clusters
            centroid_indices = self._get_centroids(fps)
            clusters = { i: [] for i in range(len(centroid_indices)) }

            # Cluster molecules based on centroids
            for i, fp in enumerate(fps):
                similarities = [DataStructs.FingerprintSimilarity(fp, fps[j]) for j in centroid_indices]
                clusters[np.argmax(similarities)].append(i)

            return clusters
    
        @abstractmethod
        def _get_centroids(self, fps : list) -> list:
            pass

class MaxMinClustering(DissimilarityClustering):

    """
    Cluster a list of SMILES strings based on molecular dissimilarity using the MaxMin algorithm.

    Attributes
    ----------
    fp_calculator : Callable, optional. 
        Function to compute molecular fingerprints.
    n_clusters : int, optional
        Number of clusters.
    seed : int, optional
        Random seed.
    """

    def __init__(
            self, 
            fp_calculator : Callable = GetMorganGenerator(radius=3, fpSize=2048),
            n_clusters : int = None,
            seed : int = 42,
        ) -> None:
        super().__init__(fp_calculator)
        self.n_clusters = n_clusters
        self.seed = seed

    def _get_centroids(self, fps : list) -> list:

        """
        Get cluster centroids using the MaxMin algorithm.
        
        Parameters
        ----------
        fps : list
            List of molecular fingerprints.
        
        Returns
        -------
        centroid_indices : list
            List of indices of cluster centroids.
        """

        self._set_n_clusters(len(fps))

        picker = rdSimDivPickers.MaxMinPicker()
        centroid_indices = picker.LazyBitVectorPick(fps, len(fps), self.n_clusters, seed=self.seed)

        return centroid_indices
    
class LeaderPickerClustering(DissimilarityClustering):

    """
    Cluster a list of SMILES strings based on molecular dissimilarity using LeadPicker to select centroids.

    Attributes
    ----------
    fp_calculator : Callable, optional.
        Function to compute molecular fingerprints.
    similarity_threshold : float, optional.
        Similarity threshold for clustering.
    """

    def __init__(
            self, 
            fp_calculator: Callable = GetMorganGenerator(radius=3, fpSize=2048),
            similarity_threshold : int = 0.7
     ) -> None:
        super().__init__(fp_calculator)
        self.similarity_threshold = similarity_threshold

    def _get_centroids(self, fps : list) -> list:

        """
        Get cluster centroids using LeadPicker.

        Parameters
        ----------
        fps : list
            List of molecular fingerprints.
        
        Returns
        -------
        centroid_indices : list
            List of indices of cluster centroids.
        """

        picker = rdSimDivPickers.LeaderPicker()
        centroid_indices = picker.LazyBitVectorPick(fps, len(fps), self.similarity_threshold)

        return centroid_indices



import os
import argparse
import pandas as pd
from timeit import default_timer as timer
from .logs.config import enable_file_logger
from .split import GloballyBalancedSplit
from .clustering import RandomClustering, LeaderPickerClustering, MaxMinClustering, MurckoScaffoldClustering

def main():

    # Parse command line arguments ###############################
    parser = argparse.ArgumentParser(formatter_class=argparse.ArgumentDefaultsHelpFormatter)

    parser.add_argument('-i', '--input', type=str, required=True,
                        help='Input file with the data in a pivoted csv/tsv format. \
                            A column with the SMILES must be provided and each target must be in a separate column.')
    parser.add_argument('-o', '--output', type=str, default=None,
                        help='Output file with the data with additional columns \
                            the assigned subset (and Minimum interset Tanimoto distance).')
    parser.add_argument('-sc','--smiles_column', type=str, default='SMILES',
                        help='Name of the column with the SMILES')
    parser.add_argument('-tc','--target_columns', type=str, nargs='+', default=None,
                        help="Name of the columns with the targets. If not provided, \
                            all columns except the SMILES and --ignore_columns' columns will be used")
    parser.add_argument('-ic','--ignore_columns', type=str, nargs='+', default=None,
                        help='Name of the columns to ignore')
    parser.add_argument('-ns', '--n_splits', type=int, default=1,
                         help='Number of splits to generate')
    parser.add_argument('-c','--clustering', type=str, default='dissimilarity_leader',
                        help='Clustering algorithm to use. \
                            Options: random, dissimilarity_leader, dissimilarity_maxmin or murcko')
    parser.add_argument('-nc','--n_clusters', type=int, default=None,
                        help='Number of clusters to use. Only used for random clustering. \
                            If None, the number of clusters is equal to the number of molecules divided by 100')
    parser.add_argument('-rs','--random_seed', type=int, default=42,
                        help='Seed for the random clustering')
    parser.add_argument('-ct','--cluster_threshold', type=float, default=0.7,
                        help='Minimum distance between cluster centers. Only used for dissimilarity clustering.')
    parser.add_argument('-s', '--sizes', type=float, nargs='+', default=[0.8, 0.1, 0.1],
                        help='Sizes of the subsets')
    parser.add_argument('-t','--time_limit', type=int, default=60,
                        help='Time limit for linear combination of clusters in seconds')
    parser.add_argument('-mtd', '--min_Tanimoto_distance', action='store_true',
                        help='Compute the minimum Tanimoto distance between a molecule in a subset and the molecules in the other subsets')
    
    # Start the timer
    start_time = timer()
    
    # Parse arguments 
    args = parser.parse_args()
    # Read input data from csv/tsv file ##########################
    if '.csv' in args.input:
        df = pd.read_csv(args.input)
    elif '.tsv' in args.input:
        df = pd.read_csv(args.input, sep='\t')
    else:
        raise ValueError('Input file must be a csv or tsv file')
    
    if args.target_columns is None and args.ignore_columns is None:
        args.target_columns = [col for col in df.columns if col != args.smiles_column]
    elif args.target_columns is None and args.ignore_columns is not None:
        args.target_columns = [col for col in df.columns if col != args.smiles_column and col not in args.ignore_columns]

    if not args.output:
        args.output = args.input.split('.')[0] 

    # Enable logging #############################################
    logSettings = enable_file_logger(
        os.path.dirname(args.output),
        "gbmtsplits.log",
        False,
        __name__,
        vars(args),
        disable_existing_loggers=False,
    )
    log = logSettings.log
    

    # Setup splitter #############################################
    if args.clustering == 'random':
        clustering = RandomClustering(n_clusters=args.n_clusters, seed=args.random_seed)
        args.output += '_RGBS.'
    elif args.clustering == 'dissimilarity_leader':
        clustering = LeaderPickerClustering(similarity_threshold=args.cluster_threshold)
        args.output += '_LPDGBS.'
    elif args.clustering == 'dissimilarity_maxmin':
        clustering = MaxMinClustering(n_clusters=args.n_clusters, seed=args.random_seed)
        args.output += '_MMDGBS.'
    elif args.clustering == 'murcko':
        clustering = MurckoScaffoldClustering()
        args.output += '_MSGBS.'
    else:
        raise ValueError('Clustering algorithm not recognized')

    splitter = GloballyBalancedSplit(
        sizes = args.sizes,
        clustering_method=clustering,
        time_limit_seconds=args.time_limit,
        min_distance=args.min_Tanimoto_distance,
        n_splits=args.n_splits
        )
    
    # Split data #################################################
    df = splitter(df, targets=args.target_columns, smiles_column=args.smiles_column)

    # Write output ###############################################
    
    # Use same extension  and compression as input file
    args.output += '.'.join(args.input.split('.')[1:])

    if '.csv' in args.output:
        df.to_csv(args.output, index=False)
    elif '.tsv' in args.output:
        df.to_csv(args.output, sep='\t', index=False)
    
    # Print elapsed time #########################################
    elapsed_time = timer() - start_time
    log.info('Elapsed time: {:.2f} seconds'.format(elapsed_time))
        
if __name__ == '__main__':

    main()


import pandas as pd
import numpy as np
import seaborn as sns
import matplotlib.pyplot as plt

from rdkit import Chem
from rdkit.Chem import AllChem

from sklearn.decomposition import PCA
from sklearn.manifold import TSNE

from typing import List

def compute_target_balance(df : pd.DataFrame, targets : List[str], subset_col : str ='Subset'):
    """
    Compute balance of subsets for each target.
    
    Parameters
    ----------
    df : pd.DataFrame
        Dataframe with a column per target and a column with the subset number.
    targets : List[str]
        List of target columns.
    subset_col : str, optional
        Name of the column with the subset number, by default 'Subset'
    
    Returns
    -------
    pd.DataFrame of shape (len(targets), len(subsets)+1)
        Dataframe with a row per target and columns containing the fraction of compounds in each subset.
    """

    subsets = sorted(df[subset_col].unique())
    df_balance = pd.DataFrame(columns=['Task']+subsets, index=range(len(targets))) 

    for i, target in enumerate(targets):
        df_target = df[[target, subset_col]].dropna()
        df_balance.loc[i, 'Task'] = target
        ntarget = len(df_target)
        for subset in subsets:
            df_balance.loc[i, subset] = len(df_target[df_target[subset_col] == subset]) / ntarget

    return df_balance

class PlottingSingleDataset():

    """Class to plot split/subset properties for a single dataset."""

    def __init__(self, data : pd.DataFrame, smiles_col : str = 'SMILES', split : str = None, dataset : str = None, targets : List[int] = None, colors : List[str] = None):
        """
        Parameters
        ----------
        data : pd.DataFrame
            Dataframe with each row containing a compound and a column per target and a column with the subset number.
        smiles_col : str, optional
            Name of the column with the SMILES, by default 'SMILES'
        split : str, optional
            Name of the split, by default None
        dataset : str, optional
            Name of the dataset, by default None
        targets : List[int], optional
            List of target columns, by default None
        colors : List[str], optional
            List of colors for the subsets, by default None
        """

        self.data = data
        self.smiles_col = smiles_col
        self.split = split
        self.dataset = dataset
        self.subset_col = 'Subset'
        self.targets = targets if targets is not None else [ col for col in data.columns if col not in [smiles_col, 'Subset', 'Dataset', 'Split', 'MinInterSetTd'] ]
        self.colors = colors if colors is not None else sns.color_palette('colorblind')

        self.title_suffix = ''
        if self.dataset : self.title_suffix += f' - ({self.dataset})'
        if self.split : self.title_suffix += f' - ({self.split})'

    def plot_pca(self, figsize : tuple = (4, 4), fname : str = None):
        """
        Plots PCA of the dataset with a different color for each subset.
        
        Parameters
        ----------
        figsize : tuple, optional
            Size of the figure, by default (4, 4)
        fname : str, optional
            Name of the file to save the figure, by default None
        
        
        Returns
        -------
        fig, ax
            Figure and axis objects.
        """
        fps = np.array([AllChem.GetMorganFingerprintAsBitVect(Chem.MolFromSmiles(s), 3, nBits=2048) for s in self.data[self.smiles_col]])
        pca = PCA(n_components=2)
        x = pca.fit_transform(fps)
        
        fig, ax = plt.subplots(1, figsize=figsize)
        for i, subset in enumerate(sorted(self.data[self.subset_col].unique())):
            ax.scatter(x[self.data[self.subset_col] == subset, 0], x[self.data[self.subset_col] == subset, 1], label=subset, alpha=0.5, color=self.colors[i], marker='.')
        
        ax.legend(frameon=False)
        ax.set_xlabel('PC1')
        ax.set_ylabel('PC2')
        ax.set_title(f'PCA {self.title_suffix}')

        if fname is not None:
            fig.savefig(fname, dpi=300, bbox_inches='tight')

        return fig, ax
    
    def plot_tsne(self, figsize : tuple = (4, 4), fname : str = None):
        """
        Plots t-SNE of the dataset with a different color for each subset.
        
        Parameters
        ----------
        figsize : tuple, optional
            Size of the figure, by default (4, 4)
        fname : str, optional
            Name of the file to save the figure, by default None
        
        Returns
        -------
        fig, ax
            Figure and axis objects.
        """
        fps = np.array([AllChem.GetMorganFingerprintAsBitVect(Chem.MolFromSmiles(s), 3, nBits=2048) for s in self.data[self.smiles_col]])
        tsne = TSNE(n_components=2)
        x = tsne.fit_transform(fps)
        
        fig, ax = plt.subplots(1, figsize=figsize)
        for i, subset in enumerate(sorted(self.data[self.subset_col].unique())):
            ax.scatter(x[self.data[self.subset_col] == subset, 0], x[self.data[self.subset_col] == subset, 1], label=subset, alpha=0.5, color=self.colors[i], marker='.')
        
        ax.legend(frameon=False)
        ax.set_xlabel('t-SNE1')
        ax.set_ylabel('t-SNE2')
        ax.set_title(f't-SNE {self.title_suffix}')

        if fname is not None:
            fig.savefig(fname, dpi=300, bbox_inches='tight')

        return fig, ax
    
    def plot_target_balance_bars(self, figsize : tuple = None, ideal_sizes : List[float] = None, fname : str = None, **kwargs):
        """
        Plots the fraction of compounds per subset for each target as stacked bars.
        
        Parameters
        ----------
        figsize : tuple, optional
            Size of the figure, by default None
        ideal_sizes : List[float], optional
            List of ideal fractions for each subset, by default None
        fname : str, optional
            Name of the file to save the figure, by default None
        
        Returns
        -------
        fig, ax
            Figure and axis objects.
        """
        df_balance = compute_target_balance(self.data, self.targets, subset_col=self.subset_col)

        if figsize is None: figsize = ( len(self.targets)/4, 4)

        fig, ax = plt.subplots(1, figsize=figsize)
        df_balance.plot.bar(x='Task', ax=ax, stacked=True, color=self.colors, width=0.8)

        if ideal_sizes is not None:
            cum = 0
            for i, size in enumerate(ideal_sizes[:-1]):
                cum += size	    
                ax.axhline(cum, color='k', linestyle='--')

        ax.set_title(f'Target balance {self.title_suffix}')
        ax.set_ylabel('Fraction of compounds')
        ax.set_xlabel('Target')
        ax.legend(frameon=False, bbox_to_anchor=(1, 1))
        ax.set_ylim(0, 1)

        if fname is not None:
            fig.savefig(fname, dpi=300, bbox_inches='tight')
        
        return fig, ax
    
    def plot_target_balance_distributions(self, ideal_sizes : List[float], figsize : tuple = None, plot_function : callable = sns.boxplot, fname : str = None, **kwargs):
        """
        Plots the distribution of the fraction of compounds per subset for each target.
        
        Parameters
        ----------
        ideal_sizes : List[float]
            List of ideal fractions for each subset
        figsize : tuple, optional
            Size of the figure, by default None
        plot_function : callable, optional
            Function to use for plotting, by default sns.boxplot
        fname : str, optional
            Name of the file to save the figure, by default None
        
        Returns
        -------
        fig, ax
            Figure and axis objects.
        """

        subsets = sorted(self.data[self.subset_col].unique())
        if len(ideal_sizes) != len(subsets):
            raise ValueError('The number of ideal sizes must be equal to the number of subsets.')
        
        df_balance = compute_target_balance(self.data, self.targets, subset_col=self.subset_col)
        
        df_balance_diff = pd.DataFrame(columns=['Subset', 'Task', 'dFraction'])
        for task in df_balance['Task'].unique():
            for subset, size in zip(subsets, ideal_sizes):
                df_balance_diff = pd.concat([df_balance_diff,
                                             pd.DataFrame({
                                                    'Subset': [subset],
                                                    'Task': [task],
                                                    'dFraction': [df_balance[df_balance['Task'] == task][subset].values[0] - size]
                                                    })
                                                ],
                                                ignore_index=True
                                            )
                
        if figsize is None: figsize = ( len(self.targets)/4, 4)

        fig, ax = plt.subplots(1, figsize=figsize)
        plot_function(x='Subset', y='dFraction', data=df_balance_diff, ax=ax, palette=self.colors)

        ax.axhline(0, color='k', linestyle='--')

        ax.set_title(f'Target balance {self.title_suffix}')
        ax.set_ylabel('dFraction (observed - ideal)')
        ax.set_xlabel('Subset')
        ax.set_ylim(-0.05, 0.05)

        if fname is not None:
            fig.savefig(fname, dpi=300, bbox_inches='tight')
        
        return fig, ax

    def plot_min_tanimoto_distance_distributions(self, figsize : tuple = None, plot_function : callable = sns.boxplot, fname : str = None, **kwargs):
        """
        Plots the distribution of the minimum Tanimoto distance between compounds in different subsets.

        
        Parameters
        ----------
        figsize : tuple, optional
            Size of the figure, by default None
        plot_function : callable, optional
            Function to use for plotting, by default sns.boxplot
        fname : str, optional
            Name of the file to save the figure, by default None
        
        Returns
        -------
        fig, ax
            Figure and axis objects.
        """

        if figsize is None: figsize = (self.data[self.subset_col].nunique(), 4)
        fig, ax = plt.subplots(1, 1, figsize=figsize)

        plot_function(x=self.subset_col, y='MinInterSetTd', data=self.data, ax=ax, palette=self.colors)

        ax.set_title(f'Min. interset Tanimoto dist. {self.title_suffix}')
        ax.set_ylabel('Min. interset Tanimoto dist.')
        ax.set_xlabel('Subset')
        ax.set_ylim(0, 1)

        if fname is not None:
            fig.savefig(fname, dpi=300, bbox_inches='tight')

        return fig, ax

    def plot_all(self, ideal_subset_sizes : List[float] = None):
        """Plot all plots."""
        self.plot_pca()
        self.plot_tsne()
        self.plot_target_balance_bars(ideal_sizes=ideal_subset_sizes)
        self.plot_target_balance_distributions(ideal_subset_sizes)
        self.plot_min_tanimoto_distance_distributions()

class PlottingCompareDatasets():
    """
    Class for plotting properties of splits/subsets of multple datasets.
    """

    def __init__(self, data : pd.DataFrame, compare_col : str, dataset_names : List[str] = None, smiles_col : str = 'SMILES', targets : List[int] = None, colors : List[str] = None):
        """
        Parameters
        ----------
        data : pd.DataFrame
            Dataframe containing the data.
        compare_col : str
            Column containing the name of the different datasets.
        dataset_names : List[str], optional
            List of names of the datasets. If None, all unique values in compare_col are used, by default None
        smiles_col : str, optional
            Name of the column containing the SMILES, by default 'SMILES'
        targets : List[int], optional
            List of targets. If None, all columns except smiles_col, compare_col, 'Subset', 'Dataset', 'Split', 'MinInterSetTd' are used, by default None
        colors : List[str], optional
            List of colors, by default None
        """
        self.data = data
        self.compare_col = compare_col
        self.dataset_names = dataset_names if dataset_names is not None else sorted(data[compare_col].unique())
        self.smiles_col = smiles_col
        self.subset_col = 'Subset'
        self.targets = targets if targets is not None else [ col for col in data.columns if col not in [smiles_col, compare_col, 'Subset', 'Dataset', 'Split', 'MinInterSetTd'] ]
        self.colors = colors if colors is not None else sns.color_palette('colorblind')

    def plot_pca(self, figsize : tuple = None, fname : str = None, **kwargs):
        """
        Plots a PCA of each dataset in subfigures with a different color for each subset.
        
        Parameters
        ----------
        figsize : tuple, optional
            Size of the figure, by default (4, 4)
        fname : str, optional
            Name of the file to save the figure, by default None
        
        Returns
        -------
        fig, ax
            Figure and axis objects.
        """
        
        fps = np.array([AllChem.GetMorganFingerprintAsBitVect(Chem.MolFromSmiles(s), 3, nBits=2048) for s in self.data[self.smiles_col]])
        
        # Fit PCA on unique compounds
        pca = PCA(n_components=2)
        unique_idx = self.data[self.smiles_col].drop_duplicates().index
        pca.fit(fps[unique_idx])

        # Transform all compounds
        x = pca.fit_transform(fps)

        if figsize is None: figsize = (len(self.dataset_names)*4, 4)
        
        fig, ax = plt.subplots(1, len(self.dataset_names), figsize=figsize)
        for i, dataset in enumerate(self.dataset_names):
            for j, subset in enumerate(sorted(self.data.Subset.unique())):
                ax[i].scatter(x[(self.data[self.compare_col] == dataset) & (self.data.Subset == subset), 0], 
                              x[(self.data[self.compare_col] == dataset) & (self.data.Subset == subset), 1], 
                              label=subset, alpha=0.5, color=self.colors[j], marker='.')
                ax[i].set_xlabel('PC1')
                ax[i].set_title(dataset)  
                ax[i].legend(frameon=False)
                
        ax[0].set_ylabel('PC2')  

        if fname is not None:
            fig.savefig(fname, dpi=300, bbox_inches='tight')

        return fig, ax
    
    def plot_tsne(self, figsize : tuple = None, fname : str = None, **kwargs):
        """
        Plots a T-SNE of each dataset in subfigures with a different color for each subset.
        
        Parameters
        ----------
        figsize : tuple, optional
            Size of the figure, by default (4, 4)
        fname : str, optional
            Name of the file to save the figure, by default None
        
        Returns
        -------
        fig, ax
            Figure and axis objects.
        """

        fps = np.array([AllChem.GetMorganFingerprintAsBitVect(Chem.MolFromSmiles(s), 3, nBits=2048) for s in self.data[self.smiles_col]])

        # Fit t-SNE on unique compounds
        tsne = TSNE(n_components=2)
        unique_idx = self.data[self.smiles_col].drop_duplicates().index
        tsne.fit(fps[unique_idx])

        # Transform all compounds
        x = tsne.fit_transform(fps)

        if figsize is None: figsize = (len(self.dataset_names)*4, 4)

        fig, ax = plt.subplots(1, len(self.dataset_names), figsize=figsize)
        for i, dataset in enumerate(self.dataset_names):
            for j, subset in enumerate(sorted(self.data.Subset.unique())):
                ax[i].scatter(x[(self.data[self.compare_col] == dataset) & (self.data.Subset == subset), 0], 
                              x[(self.data[self.compare_col] == dataset) & (self.data.Subset == subset), 1], 
                              label=subset, alpha=0.5, color=self.colors[j], marker='.')
                ax[i].set_xlabel('t-SNE 1')
                ax[i].set_title(dataset)  
                ax[i].legend(frameon=False)

        ax[0].set_ylabel('t-SNE 2')

        if fname is not None:
            fig.savefig(fname, dpi=300, bbox_inches='tight')

        return fig, ax
    
    def plot_target_balance_bars(self, figsize : tuple = None, ideal_sizes : List[float] = None, drop_task_names : bool = False, fname : str = None, **kwargs):
        """
        Plots the fraction of compound per subset for each target of each dataset in subfigures as stacked bars.
        
        Parameters
        ----------
        figsize : tuple, optional
            Size of the figure, by default None
        ideal_sizes : List[float], optional
            List of ideal sizes for each subset, by default None
        fname : str, optional
            Name of the file to save the figure, by default None
        
        Returns
        -------
        fig, ax
            Figure and axis objects.
        """

        if figsize is None: figsize = (len(self.dataset_names)*4, 4)

        fig, ax = plt.subplots(1, len(self.dataset_names), figsize=figsize, sharey=True, sharex=True)
        for i, dataset in enumerate(self.dataset_names):

            df_balance = compute_target_balance(self.data[self.data[self.compare_col] == dataset], self.targets, self.subset_col)
            df_balance.plot.bar(x='Task', ax=ax[i], stacked=True, color=self.colors, width=0.8, legend=False)

            if ideal_sizes is not None:
                cum = 0
                for size in ideal_sizes[:-1]:
                    cum += size	    
                    ax[i].axhline(cum, color='k', linestyle='--')
            
            ax[i].set_title(dataset)
            if drop_task_names:
                ax[i].set_xticklabels([])

        ax[0].set_ylabel('Fraction of compounds')
        ax[0].set_ylim(0, 1)

        # Add legend only to the last plot
        handles, labels = ax[-1].get_legend_handles_labels()
        fig.legend(handles, labels, loc='lower center', ncol=len(self.data.Subset.unique()), frameon=False)

        if fname is not None:
            fig.savefig(fname, dpi=300, bbox_inches='tight')

        return fig, ax 
    
    def plot_target_balance_distributions(self, ideal_sizes : List[float], figsize : tuple = None, plot_function : callable = sns.boxplot, fname : str = None, **kwargs):

        """ 
        Plots the distribution of the fraction of compounds per subset for each target of each dataset.
        
        Parameters
        ----------
        ideal_sizes : List[float]
            List of ideal sizes for each subset.
        figsize : tuple, optional
            Size of the figure, by default None
        plot_function : callable, optional
            Function to use for plotting, by default sns.boxplot
        fname : str, optional
            Name of the file to save the figure, by default None
        
        Returns
        -------
        fig, ax
            Figure and axis objects.
        """

        subsets = sorted(self.data[self.subset_col].unique())
        if len(ideal_sizes) != len(subsets):
            raise ValueError('The number of ideal sizes must be equal to the number of subsets.')

        if figsize is None: figsize = (self.data[self.subset_col].nunique() * len(self.dataset_names), 4)
        
        df_balance_diff = pd.DataFrame(columns=[self.compare_col, 'Task', 'Subset', 'dFraction'])
        for dataset in self.dataset_names:
            df_balance = compute_target_balance(self.data[self.data[self.compare_col] == dataset], self.targets, self.subset_col)
            for task in df_balance['Task'].unique():
                for subset, size in zip(subsets, ideal_sizes):
                    df_balance_diff = pd.concat([df_balance_diff,
                                                pd.DataFrame({
                                                        self.compare_col: [dataset],
                                                        'Subset': [subset],
                                                        'Task': [task],
                                                        'dFraction': [df_balance[df_balance['Task'] == task][subset].values[0] - size]
                                                        })
                                                    ],
                                                    ignore_index=True
                                                )
                    
        fig, ax = plt.subplots(1, 1, figsize=figsize)
        plot_function(x=self.subset_col, y='dFraction', hue=self.compare_col, hue_order=self.dataset_names, data=df_balance_diff, ax=ax, palette=self.colors, **kwargs)

        ax.axhline(0, color='k', linestyle='--')
        ax.set_ylabel('dFraction (observed - ideal)')
        ax.set_xlabel('Subset')
        ax.set_ylim(-0.05, 0.05)


        handles, labels = ax.get_legend_handles_labels()
        ax.legend([],[], frameon=False)
        fig.legend(handles, labels, loc='upper center', ncol=len(self.data.Subset.unique()), frameon=False)

        if fname is not None:
            fig.savefig(fname, dpi=300, bbox_inches='tight')


    
    def plot_min_tanimoto_distance_distributions(self, figsize : tuple = None, plot_function : callable = sns.boxplot, fname : str = None, **kwargs):
        """
        Plots the distribution of the minimum Tanimoto distance between compounds in different subsets.
        
        Parameters
        ----------
        figsize : tuple, optional
            Size of the figure, by default None
        plot_function : callable, optional
            Function to use for plotting, by default sns.boxplot
        fname : str, optional
            Name of the file to save the figure, by default None
        
        Returns
        -------
        fig, ax
            Figure and axis objects.
        """

        if figsize is None: figsize = (self.data[self.subset_col].nunique() * len(self.dataset_names), 4)
        fig, ax = plt.subplots(1, 1, figsize=figsize)

        plot_function(x=self.subset_col, y='MinInterSetTd', hue=self.compare_col, hue_order=self.dataset_names, data=self.data, ax=ax, palette=self.colors, **kwargs)

        ax.set_ylabel('Min. interset Tanimoto dist.')
        ax.set_xlabel('Subset')
        ax.set_ylim(0, 1)

        handles, labels = ax.get_legend_handles_labels()
        ax.legend([],[], frameon=False)
        fig.legend(handles, labels, loc='upper center', ncol=len(self.data.Subset.unique()), frameon=False)

        if fname is not None:
            fig.savefig(fname, dpi=300, bbox_inches='tight')

        return fig, ax

    def plot_all(self, ideal_subset_sizes : List[float] = None):
        """Plot all plots."""
        self.plot_pca()
        self.plot_tsne()
        self.plot_target_balance_bars(ideal_sizes=ideal_subset_sizes)
        self.plot_target_balance_distributions(ideal_sizes=ideal_subset_sizes)
        self.plot_min_tanimoto_distance_distributions()


import logging
import sys

logger = None

if not logger:
    logger = logging.getLogger("gbmtsplits")
    logger.setLevel(logging.INFO)


def setLogger(log):
    sys.modules[__name__].gbmtsplits = log